#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2020- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

from ibm_watson_machine_learning.helpers.connections.connections import (
    DataConnection, S3Connection, S3Location, FSLocation, AssetLocation, CP4DAssetLocation,
    WMLSAssetLocation, CloudAssetLocation, WSDAssetLocation, DeploymentOutputAssetLocation,
    NFSConnection, NFSLocation, ConnectionAsset, ConnectionAssetLocation, DatabaseLocation, ContainerLocation)
