#  -----------------------------------------------------------------------------------------
#  (C) Copyright IBM Corp. 2017- 2023.
#  https://opensource.org/licenses/BSD-3-Clause
#  -----------------------------------------------------------------------------------------

PYSPARK = 'pyspark'
SCIKIT = 'scikit-learn'
PANDAS = 'pandas'
XGBOOST= 'xgboost'
MLPIPELINE = 'mlpipelinepy'
IBMSPARKPIPELINE = 'ibmsparkpipeline'
TENSORFLOW = 'tensorflow'
HYBRID = 'hybrid'
SPSS_MODELER = 'spss-modeler'
PMML = 'pmml'
CAFFE = 'caffe'
CAFFE2 = 'caffe2'
PYTORCH = 'pytorch'
BLUECONNECT = 'blueconnect'
TORCH = 'torch'
MXNET = 'mxnet'
THEANO = 'theano'
DARKNET = 'darknet'

DISPLAY_PYSPARK = 'PySpark'
DISPLAY_SCIKIT = 'scikit-learn'
DISPLAY_PANDAS = 'Pandas'
DISPLAY_XGBOOST = 'XGBoost'
DISPLAY_MLPIPELINE = 'MLPipeline'
DISPLAY_IBMSPARKPIPELINE = 'IBMSparkPipeline'
DISPLAY_TENSORFLOW = 'Tensorflow'
DISPLAY_PMML = 'PMML'
DISPLAY_CAFFE = 'Caffe'
DISPLAY_CAFFE2 = 'Caffe2'
DISPLAY_PYTORCH = 'Pytorch'
DISPLAY_BLUECONNECT = 'Blueconnect'
DISPLAY_TORCH = 'Torch'
DISPLAY_MXNET = 'Mxnet'
DISPLAY_THEANO = 'Theano'
DISPLAY_DARKNET = 'Darknet'